#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <stdint.h>
#include <assert.h>
#include <imago2.h>

void print_usage(const char *bin);
int process(const char *out_fname, const char *img_fname, const char *mask_fname);
int upscale(struct img_pixmap *img, int xsz, int ysz);
int double_horiz(struct img_pixmap *img);
int double_vert(struct img_pixmap *img);
int is_pow2(int x);

int main(int argc, char **argv)
{
	int i;
	char *out_fname, *suffix;
	char *first = 0;

	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-') {
			if(strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "-help") == 0 || strcmp(argv[i], "--help") == 0) {
				print_usage(argv[0]);
				return 0;

			} else {
				fprintf(stderr, "invalid option: %s\n", argv[i]);
				print_usage(argv[0]);
				return 0;
			}

		} else {

			char *image_fname, *mask_fname;

			if(!first) {
				first = argv[i];
				continue;
			}

			if(!argv[i]) {
				fprintf(stderr, "must specify the images in pairs (image and mask)\n");
				return 1;
			}

			image_fname = first;
			mask_fname = argv[i];

			out_fname = malloc(strlen(image_fname) + 16);
			strcpy(out_fname, image_fname);

			if((suffix = strrchr(out_fname, '.'))) {
				*suffix++ = 0;
			}
			strcat(out_fname, "-layer.png");

			if(process(out_fname, image_fname, mask_fname) == -1) {
				return 1;
			}

			free(out_fname);
			first = 0;
		}
	}

	return 0;
}

void print_usage(const char *bin)
{
	printf("usage: %s [options] <image 0> <mask 0> <image 1> <mask 1> ... <image n> <mask n>\n", bin);
	printf("options:\n");
	printf(" -h, -help, --help: print usage and exit\n");
}

int process(const char *out_fname, const char *img_fname, const char *mask_fname)
{
	int i, j, xsz, ysz, result = -1;
	struct img_pixmap img, mask;
	uint32_t *src, *dest;

	printf("constructing %s from image %s and mask %s ...\n", out_fname, img_fname, mask_fname);

	img_init(&img);
	img_init(&mask);

	if(img_load(&img, img_fname) == -1 || img_load(&mask, mask_fname) == -1) {
		goto err;
	}

	/* make sure they're both power of 2 */
	if(!is_pow2(img.width) || !is_pow2(img.height)) {
		fprintf(stderr, "%s is not a power-of-2 image (%dx%d)\n", img_fname, img.width, img.height);
		goto err;
	}
	if(!is_pow2(mask.width) || !is_pow2(mask.height)) {
		fprintf(stderr, "%s is not a power-of-2 image (%dx%d)\n", mask_fname, mask.width, mask.height);
		goto err;
	}

	/* convert them to RGBA */
	img_convert(&img, IMG_FMT_RGBA32);
	img_convert(&mask, IMG_FMT_RGBA32);

	/* upscale the smallest one */
	xsz = img.width > mask.width ? img.width : mask.width;
	ysz = img.height > mask.height ? img.height : mask.height;

	if(img.width < xsz || img.height < ysz) {
		printf("upscaling %s: %dx%d -> %dx%d\n", img_fname, img.width, img.height, xsz, ysz);
		if(upscale(&img, xsz, ysz) == -1) {
			goto err;
		}
	}
	if(mask.width < xsz || mask.height < ysz) {
		printf("upscaling %s: %dx%d -> %dx%d\n", mask_fname, mask.width, mask.height, xsz, ysz);
		if(upscale(&mask, xsz, ysz) == -1) {
			goto err;
		}
	}

	/* move the red channel of mask to the alpha channel of img */
	src = mask.pixels;
	dest = img.pixels;

	for(i=0; i<ysz; i++) {
		for(j=0; j<xsz; j++) {
			uint32_t alpha = *src++ & 0xff;		/* keep the low 8bits of src pixel (red) */
			uint32_t pixel = *dest & 0xffffff;	/* keep the low 24bits of dest pixel (rgb) */
			*dest++ = pixel | (alpha << 24);	/* move alpha to its place and combine with rgb */
		}
	}

	if(img_save(&img, out_fname) == -1) {
		fprintf(stderr, "failed to write %s\n", out_fname);
		goto err;
	}

	result = 0;
err:
	img_destroy(&img);
	img_destroy(&mask);
	return result;
}

int upscale(struct img_pixmap *img, int xsz, int ysz)
{
	while(img->width < xsz) {
		int prev_xsz = img->width;

		if(double_horiz(img) == -1) {
			fprintf(stderr, "failed to resize image %dx%d->%dx%d\n", prev_xsz, img->height, prev_xsz * 2, img->height);
			return -1;
		}
	}

	while(img->height < ysz) {
		int prev_ysz = img->height;

		if(double_vert(img) == -1) {
			fprintf(stderr, "failed to resize image %dx%d->%dx%d\n", img->width, prev_ysz, img->height, prev_ysz * 2);
			return -1;
		}
	}

	assert(img->width == xsz && img->height == ysz);
	return 0;
}

static uint32_t mean_color(uint32_t c1, uint32_t c2)
{
	unsigned int r = ((c1 & 0xff) + (c2 & 0xff)) / 2;
	unsigned int g = (((c1 >> 8) & 0xff) + ((c2 >> 8) & 0xff)) / 2;
	unsigned int b = (((c1 >> 16) & 0xff) + ((c2 >> 16) & 0xff)) / 2;
	unsigned int a = (((c1 >> 24) & 0xff) + ((c2 >> 24) & 0xff)) / 2;

	return r | (g << 8) | (b << 16) | (a << 24);
}

int double_horiz(struct img_pixmap *img)
{
	int i, j;
	int xsz = img->width;
	int ysz = img->height;
	int newxsz = xsz * 2;
	uint32_t *newpixels, *src, *dest;
	unsigned int wrapmask = xsz - 1;	/* assuming pow-2 */

	if(!(newpixels = malloc(newxsz * img->height * 4))) {
		return -1;
	}

	src = img->pixels;
	dest = newpixels;

	for(i=0; i<ysz; i++) {
		for(j=0; j<xsz; j++) {
			uint32_t next = src[(j + 1) & wrapmask];
			uint32_t pix = src[j];

			*dest++ = pix;
			*dest++ = mean_color(pix, next);
		}
		src += xsz;
	}

	img->width = newxsz;
	free(img->pixels);
	img->pixels = newpixels;
	return 0;
}

int double_vert(struct img_pixmap *img)
{
	int i, j;
	int xsz = img->width;
	int ysz = img->height;
	int newysz = ysz * 2;
	uint32_t *newpixels, *src, *dest;

	if(!(newpixels = malloc(img->width * newysz * 4))) {
		return -1;
	}

	src = img->pixels;
	dest = newpixels;

	for(i=0; i<ysz; i++) {
		for(j=0; j<xsz; j++) {
			uint32_t next = (i + 1) < ysz ? src[xsz] : ((uint32_t*)img->pixels)[j];
			uint32_t pix = src[0];

			dest[0] = pix;
			dest[xsz] = mean_color(pix, next);

			++src;
			++dest;
		}
		dest += xsz;
	}

	img->height = newysz;
	free(img->pixels);
	img->pixels = newpixels;
	return 0;
}

int is_pow2(int x)
{
	int i, nbits = sizeof x / CHAR_BIT;
	int nset = 0;

	for(i=0; i<nbits; i++) {
		if(x & (1 << i)) {
			if(nset) return 0;
			nset++;
		}
	}
	return 1;
}
