#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <vector>

#include "opengl.h"

#ifdef WIN32
#include <windows.h>
#else
#include <unistd.h>
#endif

#include "terrain.h"
#include "imtk/imtk.h"
#include "sdr.h"

static void disp();
static void gui();
static void idle();
static void limit_framerate(int fps);
static void reshape(int x, int y);
static void keyb(unsigned char key, int x, int y);
static void mouse(int bn, int st, int x, int y);
static void motion(int x, int y);
static void passive_motion(int x, int y);
static void picking(int x, int y);
static Ray calc_picking_ray(int x, int y);
static bool parse_cmdline(int argc, char **argv);

static int win_width, win_height;
static float cam_theta = 25.0, cam_phi, cam_dist = 6.0;
static Terrain terrain;
static bool show_normals, show_wireframe;
static float normscale = 0.1;
static int usubdiv, vsubdiv;
static bool use_hit_pos;
static Vector3 hit_pos;

static bool show_movement, auto_move;
static float move_v;

static const char *opt_hmap_fname = "hmap.png";
static std::vector<const char*> opt_layer_fnames;
static std::vector<float> opt_layer_tile;

static unsigned int sdr;


int main(int argc, char **argv)
{
	glutInitWindowSize(1280, 800);
	glutInit(&argc, argv);

	if(!parse_cmdline(argc, argv)) {
		return 1;
	}

	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutCreateWindow("terrain");

	glutDisplayFunc(disp);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyb);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutPassiveMotionFunc(passive_motion);
	glutIdleFunc(idle);

	glewInit();

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	if(!terrain.load(opt_hmap_fname)) {
		return 1;
	}
	terrain.set_scale(5.0);
	terrain.set_height_scale(1.0);
	usubdiv = terrain.get_map_width() - 1;
	vsubdiv = terrain.get_map_height() - 1;

	for(size_t i=0; i<opt_layer_fnames.size(); i++) {
		Texture *tex = new Texture;
		if(!tex->load(opt_layer_fnames[i])) {
			delete tex;
			return 1;
		}

		float ts = opt_layer_tile[i];
		int idx = terrain.add_layer(tex);
		terrain.get_layer(idx).tex_scale = Vector2(ts, ts);
		printf("layer %d texture scale: %f\n", (int)i, ts);
	}

	if(!(sdr = create_program_load("sdr/vertex.glsl", "sdr/pixel.glsl"))) {
		return 1;
	}
	terrain.set_shader(sdr);

	glClearColor(0.25, 0.25, 0.25, 1);
	imtk_set_color(IMTK_TEXT_COLOR, 1, 1, 1, 1);

	glutMainLoop();
	return 0;
}

static void disp()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0, 0, -cam_dist);
	glRotatef(cam_phi, 1, 0, 0);
	glRotatef(cam_theta, 0, 1, 0);

	if(show_wireframe) {
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	}
	terrain.draw(usubdiv, vsubdiv);
	if(show_wireframe) {
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}

	if(show_normals) {
		glPushAttrib(GL_ENABLE_BIT);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);
		glDisable(GL_LIGHTING);

		glColor4f(0.2, 1, 0.25, 0.5);
		terrain.draw_normals(normscale, usubdiv, vsubdiv);

		glPopAttrib();
	}

	if(show_movement) {
		const float rad = 0.05;
		const float blue[] = {0.2, 0.3, 1.0, 1.0};

		if(auto_move) {
			move_v = fmod(glutGet(GLUT_ELAPSED_TIME) / 10000.0, 1.0);
		}

		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, blue);

		// draw the moving "dots" accross the terrain
		for(int i=0; i<10; i++) {
			float u = (float)i / 9.0;

			Vector2 pos = terrain.uv_to_world(Vector2(u, move_v));
			float height = terrain.get_world_height(pos);

			glPushMatrix();
			glTranslatef(pos.x, height + rad, pos.y);
			glutSolidSphere(rad, 12, 6);
			glPopMatrix();
		}
	}

	if(use_hit_pos) {
		const float rad = 0.075;
		float red[] = {1, 0.25, 0.1, 1};
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, red);

		glPushMatrix();
		glTranslatef(hit_pos.x, hit_pos.y + rad, hit_pos.z);
		glutSolidSphere(rad, 16, 8);
		glPopMatrix();
	}


	gui();

	glutSwapBuffers();
	assert(glGetError() == GL_NO_ERROR);

	limit_framerate(40);
}


static void gui()
{
	imtk_begin();
	imtk_layout_start(10, 0);
	imtk_layout_spacing(10);
	imtk_layout_dir(IMTK_VERTICAL);

	char buf[128];
	sprintf(buf, "subdiv (change with +/-): %dx%d\n", usubdiv, vsubdiv);
	imtk_label(buf, IMTK_AUTO, IMTK_AUTO);

	imtk_layout_dir(IMTK_HORIZONTAL);
	imtk_label("terrain scale:", IMTK_AUTO, IMTK_AUTO);
	terrain.set_scale(imtk_slider(IMUID, terrain.get_scale(), 0.1, 20.0, IMTK_AUTO, IMTK_AUTO));
	imtk_layout_dir(IMTK_VERTICAL);

	imtk_layout_dir(IMTK_HORIZONTAL);
	imtk_label("terrain height:", IMTK_AUTO, IMTK_AUTO);
	terrain.set_height_scale(imtk_slider(IMUID, terrain.get_height_scale(), 0.1, 10.0, IMTK_AUTO, IMTK_AUTO));
	imtk_layout_dir(IMTK_VERTICAL);

	if((show_normals = imtk_checkbox(IMUID, "normals", IMTK_AUTO, IMTK_AUTO, show_normals))) {
		imtk_layout_dir(IMTK_HORIZONTAL);
		imtk_label("normal scale:", IMTK_AUTO, IMTK_AUTO);
		normscale = imtk_slider(IMUID, normscale, 0.01, 2.0, IMTK_AUTO, IMTK_AUTO);
		imtk_layout_dir(IMTK_VERTICAL);
	}
	show_wireframe = imtk_checkbox(IMUID, "wireframe", IMTK_AUTO, IMTK_AUTO, show_wireframe);

	if((show_movement = imtk_checkbox(IMUID, "moving dots", IMTK_AUTO, IMTK_AUTO, show_movement))) {
		if((auto_move = imtk_checkbox(IMUID, "auto movement", IMTK_AUTO, IMTK_AUTO, auto_move))) {
			// automatic movement, just show a progress bar here
			imtk_progress(IMUID, move_v, IMTK_AUTO, IMTK_AUTO);
		} else {
			// provide a slider
			imtk_layout_dir(IMTK_HORIZONTAL);
			imtk_label("movement:", IMTK_AUTO, IMTK_AUTO);
			move_v = imtk_slider(IMUID, move_v, 0.0, 1.0, IMTK_AUTO, IMTK_AUTO);
			imtk_layout_dir(IMTK_VERTICAL);
		}
	}

	imtk_end();
}

static void idle()
{
	glutPostRedisplay();
}

static void limit_framerate(int fps)
{
	static long prev_upd;
	long msec = glutGet(GLUT_ELAPSED_TIME);
	long dt = msec - prev_upd;

	long interval = 1000 / fps;
	long leftover = interval - dt;
	if(leftover > 1) {
#ifdef WIN32
		Sleep(leftover - 1);
#else
		usleep((leftover - 1) * 1000);
#endif
	}

	prev_upd = glutGet(GLUT_ELAPSED_TIME);
}


static void reshape(int x, int y)
{
	win_width = x;
	win_height = y;

	glViewport(0, 0, x, y);
	imtk_set_viewport(x, y);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(50.0, (float)x / (float)y, 0.5, 500.0);
}

static void keyb(unsigned char key, int x, int y)
{
	switch(key) {
	case 27:
		exit(0);

	case 'w':
		show_wireframe = !show_wireframe;
		glutPostRedisplay();
		break;

	case 'n':
		show_normals = !show_normals;
		glutPostRedisplay();
		break;

	case '=':
		usubdiv++;
		vsubdiv++;
		glutPostRedisplay();
		break;

	case '-':
		usubdiv--;
		vsubdiv--;
		if(usubdiv < 1 || vsubdiv < 1) {
			usubdiv++;
			vsubdiv++;
		}
		glutPostRedisplay();
		break;

	default:
		break;
	}
}


static int prev_x, prev_y;
static bool bnstate[32];
static unsigned int modkeys;

static void mouse(int bn, int st, int x, int y)
{
	modkeys = glutGetModifiers();

	prev_x = x;
	prev_y = y;
	bnstate[bn - GLUT_LEFT_BUTTON] = st == GLUT_DOWN;

	// get the GUI bounding box, to avoid picking in there
	int guibox[4];
	imtk_layout_get_bounds(guibox);

	if(x <= guibox[0] || x > guibox[0] + guibox[2] || y <= guibox[1] || y > guibox[1] + guibox[3]) {
		if(!modkeys && bnstate[0]) {
			picking(x, y);
		}
	}

	imtk_inp_mouse(bn, st == GLUT_DOWN ? IMTK_DOWN : IMTK_UP);
}

static void motion(int x, int y)
{
	int dx = x - prev_x;
	int dy = y - prev_y;
	prev_x = x;
	prev_y = y;

	if(modkeys) {
		if(bnstate[0]) {
			cam_theta += dx * 0.5;
			cam_phi += dy * 0.5;

			if(cam_phi < -90) cam_phi = -90;
			if(cam_phi > 90) cam_phi = 90;
		}

		if(bnstate[2]) {
			cam_dist += dy * 0.1;
			if(cam_dist < 0.0) {
				cam_dist = 0.0;
			}
		}
	} else {

		// get the GUI bounding box, to avoid picking in there
		int guibox[4];
		imtk_layout_get_bounds(guibox);

		if(x <= guibox[0] || x > guibox[0] + guibox[2] || y <= guibox[1] || y > guibox[1] + guibox[3]) {
			if(!modkeys && bnstate[0]) {
				picking(x, y);
			}
		}
	}

	imtk_inp_motion(x, y);
	glutPostRedisplay();
}

static void passive_motion(int x, int y)
{
	imtk_inp_motion(x, y);
	glutPostRedisplay();
}

static void picking(int x, int y)
{
	Ray pick_ray = calc_picking_ray(x, win_height - y);
	float dist;
	if(terrain.intersect(pick_ray, &dist)) {
		hit_pos = pick_ray.origin + pick_ray.dir * dist;
		use_hit_pos = true;
	} else {
		use_hit_pos = false;
	}
}

static Ray calc_picking_ray(int x, int y)
{
	int viewport[4];
	double modelview[16], projection[16];
	double posx, posy, posz;

	glGetDoublev(GL_MODELVIEW_MATRIX, modelview);
	glGetDoublev(GL_PROJECTION_MATRIX, projection);
	glGetIntegerv(GL_VIEWPORT, viewport);

	gluUnProject(x, y, 0, modelview, projection, viewport, &posx, &posy, &posz);
	Vector3 pnear = Vector3(posx, posy, posz);

	gluUnProject(x, y, 1, modelview, projection, viewport, &posx, &posy, &posz);
	Vector3 pfar = Vector3(posx, posy, posz);

	return Ray(pnear, pfar - pnear);
}

static bool parse_cmdline(int argc, char **argv)
{
	float tiling = 1.0;

	for(int i=1; i<argc; i++) {
		if(argv[i][0] == '-') {
			if(argv[i][2] != 0) {
				fprintf(stderr, "invalid option: %s\n", argv[i]);
				return false;
			}
			switch(argv[i][1]) {
			case 'm':
				opt_hmap_fname = argv[++i];
				break;

			case 'l':
				opt_layer_fnames.push_back(argv[++i]);
				opt_layer_tile.push_back(tiling);
				break;

			case 't':
				if(!(tiling = atoi(argv[++i]))) {
					fprintf(stderr, "invalid tiling value: %s\n", argv[i]);
					return false;
				}
				break;

			default:
				fprintf(stderr, "invalid option: %s\n", argv[i]);
				return false;
			}
		} else {
			fprintf(stderr, "unexpected argument: %s\n", argv[i]);
			return false;
		}
	}
	return true;
}
