#include "opengl.h"
#include "terrain.h"
#include "box.h"
#include "imago2.h"

Terrain::Terrain()
{
	xsz = ysz = 0;
	dx = dy = 0;
	hmap = 0;
	scale = height_scale = 1;

	dlist = dlist_norm = 0;

	dlist_sub[0] = dlist_sub[1] = dlist_norm_sub[0] = dlist_norm_sub[1] = -1;
	dlist_norm_sz = -1.0;

	sdr = 0;
}

Terrain::~Terrain()
{
	if(hmap) {
		img_free_pixels(hmap);
	}

	glDeleteLists(dlist, 2);
}

int Terrain::add_layer()
{
	Material mtl;
	layers.push_back(mtl);
	return (int)layers.size() - 1;
}

int Terrain::add_layer(const Material &mtl)
{
	layers.push_back(mtl);
	return (int)layers.size() - 1;
}

int Terrain::add_layer(Texture *tex, int textype)
{
	Material mtl;
	mtl.tex[textype] = tex;
	layers.push_back(mtl);
	return (int)layers.size() -1;
}

bool Terrain::set_layer_texture(int idx, Texture *tex, int textype)
{
	if(idx < 0 || idx >= (int)layers.size()) {
		return false;
	}
	layers[idx].tex[textype] = tex;
	return true;
}

const Material &Terrain::get_layer(int idx) const
{
	if(idx < 0 || idx >= (int)layers.size()) {
		idx = 0;
	}
	return layers[idx];
}

Material &Terrain::get_layer(int idx)
{
	if(idx < 0 || idx >= (int)layers.size()) {
		idx = 0;
	}
	return layers[idx];
}

void Terrain::set_shader(unsigned int sdr)
{
	this->sdr = sdr;
}

unsigned int Terrain::get_shader() const
{
	return sdr;
}

void Terrain::set_scale(float s)
{
	if(s != scale) {
		scale = s;
		invalidate_mesh();
	}
}

float Terrain::get_scale() const
{
	return scale;
}

void Terrain::set_height_scale(float s)
{
	if(s != height_scale) {
		height_scale = s;
		invalidate_mesh();
	}
}

float Terrain::get_height_scale() const
{
	return height_scale;
}

bool Terrain::load(const char *fname)
{
	if(!(hmap = (float*)img_load_pixels(fname, &xsz, &ysz, IMG_FMT_GREYF))) {
		fprintf(stderr, "failed to load image: %s\n", fname);
		return false;
	}

	dx = 1.0 / (float)(xsz - 1);
	dy = 1.0 / (float)(ysz - 1);
	return true;
}

Vector2 Terrain::world_to_uv(const Vector2 &pt) const
{
	return pt  / scale + 0.5;
}

Vector2 Terrain::uv_to_world(const Vector2 &pt) const
{
	return (pt - 0.5) * scale;
}

float Terrain::lookup(int x, int y) const
{
	int px = x < 0 ? 0 : (x >= xsz ? x = xsz - 1 : x);
	int py = y < 0 ? 0 : (y >= ysz ? y = ysz - 1 : y);
	return hmap[py * xsz + px];
}

float Terrain::get_height(const Vector2 &pt) const
{
	float fxsz = (float)xsz;
	float fysz = (float)ysz;

	// the floor of x * xsz is the pixel x coordinate (e.g. [0.0, 1.0, 2.0, ... 127.0] for 128x128)
	float floorx = floor(pt.x * fxsz);
	float floory = floor(pt.y * fysz);

	// (x0, y0) is the lower left pixel in normalized coords of the four pixels surrounding pt
	float x0 = floorx / fxsz;
	float y0 = floory / fysz;

	// (u, v) is the relative position of pt within the 4pix square
	// so if it's exactly at the lower-right it'll be (0, 0), in the middle it'll be (0.5, 0.5) etc.
	float u = (pt.x - x0) * fxsz;
	float v = (pt.y - y0) * fysz;

	// we need integer pixel coordinates to use lookup()
	int px = (int)floorx;
	int py = (int)floory;

	// the heights at the 4 corners
	float h00 = lookup(px, py);
	float h10 = lookup(px + 1, py);
	float h01 = lookup(px, py + 1);
	float h11 = lookup(px + 1, py + 1);

	// first interpolate along the top and bottom edges
	float top = lerp(h01, h11, u);
	float bot = lerp(h00, h10, u);
	// then vertically between the top/bot values
	return lerp(bot, top, v);
}

Vector3 Terrain::get_normal(const Vector2 &pt) const
{
	// compute partial derivatives (slopes) in respect to X and Y using central differences
	float dfdx = get_height(Vector2(pt.x + dx, pt.y)) - get_height(Vector2(pt.x - dx, pt.y));
	float dfdy = get_height(Vector2(pt.x, pt.y + dy)) - get_height(Vector2(pt.x, pt.y - dy));

	/* we need to multiply with the final scale factors while constructing the tangent and
	 * bitangent vectors, otherwise the "aspect" of the normals will be wrong while changing
	 * the two scale factors independently.
	 */
	Vector3 tang = Vector3(2.0 * dx * scale, dfdx * height_scale, 0);
	Vector3 bitan = Vector3(0, dfdy * height_scale, 2.0 * dy * scale);
	return cross_product(bitan, tang).normalized();
}

float Terrain::get_world_height(const Vector2 &pt) const
{
	return get_height(world_to_uv(pt)) * height_scale;
}

/* step over pixel distances until we cross the terrain, then determine the
 * point of intersection
 */
bool Terrain::intersect(const Ray &ray, float *dist) const
{
	// axis-aligned bounding box of the terrain
	Box aabb;
	aabb.min = Vector3(-0.5 * scale, 0, -0.5 * scale);
	aabb.max = Vector3(0.5 * scale, height_scale, 0.5 * scale);

	Ray mray;
	mray.origin = ray.origin;
	// find a reasonable step size
	float pixsz = dx > dy ? dx : dy;
	float raysz = ray.dir.length();
	mray.dir = ray.dir / raysz * pixsz;

	float aabb_dist = 0.0;

	if(!aabb.contains(mray.origin)) {
		/* if we're not in the aabb, calculate the intersection point of the ray
		 * with the aabb, to start ray-marching from that point.
		 */
		if(!aabb.intersect(ray, &aabb_dist)) {
			return false;	// the ray misses the terrain completely
		}
		mray.origin += ray.dir * aabb_dist;
	}


	// ray-march over the terrain until we find an intersection or leave the box
	int num_steps = 0;
	do {
		float mdist;
		if(intersect_micro(mray, &mdist, 0.01)) {
			/* to calculate the distance travelled along the original ray we need to
			 * first add the distance to the aabb where we started ray marching. Then
			 * add the fraction of the last step to the number of micro-steps up to
			 * this point, which have to be converted to the original ray scale by
			 * multiplying by the magnitude of the micro-ray direction, over the
			 * magnitude of the original ray direction.
			 */
			*dist = aabb_dist + (mdist + (float)num_steps) * pixsz / raysz;
			return true;
		}

		mray.origin += mray.dir;
		num_steps++;
	} while(aabb.contains(mray.origin));

	return false;
}

bool Terrain::intersect_micro(const Ray &ray, float *dist, float thres) const
{
	Vector3 start = ray.origin;
	Vector3 end = ray.origin + ray.dir;

	float hstart = get_world_height(Vector2(start.x, start.z));
	float hend = get_world_height(Vector2(end.x, end.z));

	/* if both the start and the end of the ray are above or below the heightfield
	 * then there's no possible intersection in micro-scales.
	 */
	if(start.y > hstart && end.y > hend) {
		return false;	// all above
	}
	if(start.y < hstart && end.y < hend) {
		return false;	// all below
	}

	/* otherwise (one on one side and the other on the other side), we're straddling
	 * the heightfield. So, find the mid-point and binary-search until we find an
	 * estimated intersection point within the limits defined by "thres".
	 */
	Vector3 mid = lerp(start, end, 0.5);
	float hmid = get_world_height(Vector2(mid.x, mid.z));
	float dh = mid.y - hmid;

	int iter = 0;	// iter is there to avoid infinite loops in marginal cases
	while(fabs(dh) > thres && iter < 100) {
		if(dh > 0) {	// mid is above the surface
			start = mid;
		} else {		// mid is below the surface
			end = mid;
		}
		mid = lerp(start, end, 0.5);
		hmid = get_world_height(Vector2(mid.x, mid.z));
		dh = mid.y - hmid;

		iter++;
	}

	// found the intersection point, calculate the parametric distance and return true
	*dist = (mid - ray.origin).length() / ray.dir.length();
	return true;
}

void Terrain::draw(int usub, int vsub) const
{
	if(layers.empty()) {
		float ltgreen[] = {0.766, 0.889, 0.725, 1};
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, ltgreen);
		draw_mesh(usub, vsub);
		return;
	}

	glUseProgram(sdr);

	glPushAttrib(GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glDepthFunc(GL_LEQUAL);

	for(size_t i=0; i<layers.size(); i++) {
		layers[i].bind();
		draw_mesh(usub, vsub);
	}

	glPopAttrib();

	glUseProgram(0);
}

void Terrain::draw_mesh(int usub, int vsub) const
{
	if(usub < 1) usub = xsz - 1;
	if(vsub < 1) vsub = ysz - 1;

	if(dlist && dlist_sub[0] == usub && dlist_sub[1] == vsub) {
		glCallList(dlist);
		return;
	}

	if(!dlist) {
		dlist = glGenLists(2);
		dlist_norm = dlist + 1;
	}
	glNewList(dlist, GL_COMPILE_AND_EXECUTE);
	dlist_sub[0] = usub;
	dlist_sub[1] = vsub;

	float du = 1.0 / (float)usub;
	float dv = 1.0 / (float)vsub;

	glBegin(GL_QUADS);

	float v = 0.0;
	for(int i=0; i<vsub; i++) {
		float u = 0.0;
		for(int j=0; j<usub; j++) {

			Vector2 uv(u, v);
			Vector3 norm = get_normal(uv);
			Vector2 wpt = uv_to_world(Vector2(u, v));
			glNormal3f(norm.x, norm.y, norm.z);
			glTexCoord2f(uv.x, uv.y);
			glVertex3f(wpt.x, get_height(Vector2(u, v)) * height_scale, wpt.y);

			uv = Vector2(u, v + dv);
			norm = get_normal(uv);
			wpt = uv_to_world(uv);
			glNormal3f(norm.x, norm.y, norm.z);
			glTexCoord2f(uv.x, uv.y);
			glVertex3f(wpt.x, get_height(uv) * height_scale, wpt.y);

			uv = Vector2(u + du, v + dv);
			norm = get_normal(uv);
			wpt = uv_to_world(uv);
			glNormal3f(norm.x, norm.y, norm.z);
			glTexCoord2f(uv.x, uv.y);
			glVertex3f(wpt.x, get_height(uv) * height_scale, wpt.y);

			uv = Vector2(u + du, v);
			norm = get_normal(uv);
			wpt = uv_to_world(uv);
			glNormal3f(norm.x, norm.y, norm.z);
			glTexCoord2f(uv.x, uv.y);
			glVertex3f(wpt.x, get_height(uv) * height_scale, wpt.y);

			u += du;
		}
		v += dv;
	}
	glEnd();

	glEndList();
}

void Terrain::draw_normals(float sz, int usub, int vsub) const
{
	if(usub < 1) usub = xsz - 1;
	if(vsub < 1) vsub = ysz - 1;

	if(dlist_norm && dlist_norm_sub[0] == usub && dlist_norm_sub[1] == vsub && fabs(dlist_norm_sz - sz) < 0.0001) {
		glCallList(dlist_norm);
		return;
	}

	if(!dlist_norm) {
		dlist = glGenLists(2);
		dlist_norm = dlist + 1;
	}
	glNewList(dlist_norm, GL_COMPILE_AND_EXECUTE);
	dlist_norm_sub[0] = usub;
	dlist_norm_sub[1] = vsub;
	dlist_norm_sz = sz;

	float du = 1.0 / (float)usub;
	float dv = 1.0 / (float)vsub;

	glBegin(GL_LINES);

	float v = 0.0;
	for(int i=0; i<vsub + 1; i++) {
		float u = 0.0;
		for(int j=0; j<usub + 1; j++) {
			Vector2 uv(u, v);
			Vector2 wpt = uv_to_world(uv);
			Vector3 norm = get_normal(uv) * sz * height_scale;

			Vector3 p(wpt.x, get_height(uv) * height_scale, wpt.y);
			glVertex3f(p.x, p.y, p.z);
			glVertex3f(p.x + norm.x, p.y + norm.y, p.z + norm.z);

			u += du;
		}
		v += dv;
	}
	glEnd();

	glEndList();
}

void Terrain::invalidate_mesh()
{
	// this will force recreation of the display lists on the next draw call
	dlist_sub[0] = dlist_sub[1] = dlist_norm_sub[0] = dlist_norm_sub[1] = -1;
}
